/*************************************************************
 *
 *  writeout.c - output subroutines for dla3d
 *
 *  Copyright (C) 2000-2004  Mark J. Stock, mstock@umich.edu
 *
 *  This file is part of dla3d.
 *
 *  dla3d is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  dla3d is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with dla3d; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 *********************************************************** */


#include "structs.h"

int write_output(fileprop_ptr,sim_ptr,cell_ptr);
int write_particle_count(cell_ptr);
int write_pgm_plane(fileprop_ptr,cell_ptr,char*);
int write_pgm_density_3d(fileprop_ptr,cell_ptr,field3_ptr,char*);
int write_pgm_density_2d(fileprop_ptr,cell_ptr,field2_ptr,char*);
int put_part_into_array(cell_ptr,cell_ptr,FLOAT**,int,int);
int write_rad(sim_ptr,cell_ptr,char*);
int write_rad_cell(cell_ptr,FILE*,int);
int write_part(cell_ptr,char*);
int write_part_cell(cell_ptr,FILE*);


/*
 * Controls writing of output, once decision to write has been made
 */
int write_output(fileprop_ptr file,sim_ptr sim,cell_ptr top) {

   char outfile[80],outfile2[80],command[160];

   /* actually write the output here */
   if (file->write_dot) {
      if (file->write_pgm || file->write_gif || file->write_png) {
         sprintf(outfile,"output_%04d.pgm",sim->next_output_index);
         write_pgm_plane(file,top,outfile);
      }
      if (file->write_gif) {
         sprintf(outfile2,"output_%04d.gif",sim->next_output_index);
         sprintf(command,"ppmtogif %s > %s\n",outfile,outfile2);
         system(command);
      }
      if (file->write_png) {
         sprintf(outfile2,"output_%04d.png",sim->next_output_index);
         sprintf(command,"pnmtopng %s > %s\n",outfile,outfile2);
         system(command);
      }
      if (!file->write_pgm) {
         sprintf(command,"rm %s\n",outfile);
         system(command);
      }
   }
   if (file->write_rad) {
      zero_all_masses(top);
      sum_tipward_mass(top);
      sprintf(outfile,"outputp_%04d.rad",sim->next_output_index);
      write_rad(sim,top,outfile);
   }
   if (file->write_part) {
      sprintf(outfile,"output_%04d.part",sim->next_output_index);
      write_part(top,outfile);
   }
                                                                                 
   // create the density field, hopefully more later
   if (sim->use_density_field) {
      create_density_field_2d(top,top,sim->ff2);
      if (file->write_pgm || file->write_gif || file->write_png) {
         sprintf(outfile,"dens_%04d.pgm",sim->next_output_index);
         write_pgm_density_2d(file,top,sim->ff2,outfile);
      }
      if (file->write_gif) {
         sprintf(outfile2,"dens_%04d.gif",sim->next_output_index);
         sprintf(command,"ppmtogif %s > %s\n",outfile,outfile2);
         system(command);
      }
      if (file->write_png) {
         sprintf(outfile2,"dens_%04d.png",sim->next_output_index);
         sprintf(command,"pnmtopng %s > %s\n",outfile,outfile2);
         system(command);
      }
      if (!file->write_pgm) {
         sprintf(command,"rm %s\n",outfile);
         system(command);
      }
   }

   return(0);
}


/*
 * write a heirarchical description of the particles in cells
 */
int write_particle_count(cell_ptr cell){

   int i,j,k;

   if (cell->level == 0) fprintf(stdout,"\n");

   if (cell->num > 0) {
      for (i=-1;i<cell->level;i++) fprintf(stdout,"  ");
      fprintf(stdout,"cell at level %d has %d particles\n",cell->level,cell->num);
   }

   if (cell->has_subcells)
      for (i=0;i<2;i++)
         for (j=0;j<2;j++)
            for (k=0;k<2;k++)
               write_particle_count(cell->s[i][j][k]);

   return(0);
}


/*
 * Write a PGM image of the points projected onto the xy-plane
 */
int write_pgm_plane(fileprop_ptr file,cell_ptr cell,char *filename){

   int do_fireworks = FALSE;
   int i,j;
   int nx = file->out_img_size;
   int ny = file->out_img_size;
   FLOAT scale = 1.0;
   // FLOAT **array = (FLOAT **)malloc(nx * sizeof(FLOAT *));
   FIELD2 my_array;		// initialize struct for density field
   field2_ptr array = &my_array;
   int printval;
   FILE *outfile;

   // fprintf(stderr,"I'm here %d\n",file->out_img_size); fflush(stderr);

   // allocate space for 2D array
   // array = (FLOAT **)malloc(nx * sizeof(FLOAT *));
   // array[0] = (FLOAT *)malloc(nx * ny * sizeof(FLOAT));
   // for (i=1; i<nx; i++)
      // array[i] = array[0] + i * ny;
   // allocated once in setup.c
   // file->out->rho = allocate_2d_array_F(nx,ny);
   

   /* zero the array */
   /* Not zeroing this array must use the pervious numbers, and this
      thing just adds the new locations to the old, looks like the
      burst of a firework! */
   if (do_fireworks) {
      /* do not clear memory */
   } else {
      for (j=ny-1; j>=0; j--) {
         for (i=0; i<nx; i++) {
            file->out->rho[i][j] = 0.0;
         }
      }
   }

   /* fill in the array */
   put_part_into_array(cell,cell,file->out->rho,nx,ny);

   /* open file for writing */
   outfile = fopen(filename,"w");
   if (outfile==NULL) {
      fprintf(stderr,"Could not open output file %s\n",filename);
      fflush(stderr);
      exit(0);
   }

   /* plot an z-plane */
   fprintf(outfile,"P2\n%d %d\n%d\n",nx,ny,255);
   for (j=ny-1; j>=0; j--) {
      for (i=0; i<nx; i++) {
         printval = (int)(256*file->out->rho[i][j]*scale);
         if (printval<0) printval = 0;
         if (printval>255) printval = 255;
         fprintf(outfile,"%d\n",printval);
      }
   }

   /* close file */
   fclose(outfile);

   /* free memory from 2D array! */
   // free(array[0]);
   // free(array);

   /* if all went well, return zero */
   return(0);
}


/*
 * Write a PGM image of the density field
 */
int write_pgm_density_3d(fileprop_ptr file,cell_ptr cell,field3_ptr ff,char *filename){

   int do_middle_slice = FALSE;
   int i,j,k;
   int kindex = ff->n[1]/2;
   int nx = ff->n[0];
   int ny = ff->n[2];
   FLOAT scale = 2.0;
   FLOAT **array = (FLOAT **)malloc(nx * sizeof(FLOAT *));
   int printval;
   FILE *outfile;

   /* allocate space for 2D array */
   array[0] = (FLOAT *)malloc(nx * ny * sizeof(FLOAT));
   for (i=1; i<nx; i++)
      array[i] = array[0] + i * ny;

   /* zero the array */
   for (j=ny-1; j>=0; j--)
      for (i=0; i<nx; i++)
         array[i][j] = 0.0;

   /* fill the array */
   for (i=0; i<nx; i++)
      for (j=0; j<ny; j++) {
         if (do_middle_slice) array[i][j] = ff->rho[i][kindex][j];
         else {
            for (k=0; k<ff->n[1]; k++)
               array[i][j] += ff->rho[i][k][j];
            array[i][j] *= 0.0025;
         }
      }

   // scale the array?


   /* open file for writing */
   outfile = fopen(filename,"w");
   if (outfile==NULL) {
      fprintf(stderr,"Could not open output file %s\n",filename);
      fflush(stderr);
      exit(0);
   }

   /* plot a y-plane */
   fprintf(outfile,"P2\n%d %d\n%d\n",nx,ny,255);
   for (j=ny-1; j>=0; j--) {
      for (i=0; i<nx; i++) {
         printval = (int)(256.0*array[i][j]*scale);
         if (printval<0) printval = 0;
         if (printval>255) printval = 255;
         fprintf(outfile,"%d\n",printval);
      }
   }

   /* close file */
   fclose(outfile);

   /* free memory from 2D array! */
   free(array[0]);
   free(array);

   /* if all went well, return zero */
   return(0);
}


/*
 * Write a PGM image of the density field
 */
int write_pgm_density_2d(fileprop_ptr file,cell_ptr cell,field2_ptr ff,char *filename){

   int i,j;
   int nx = ff->n[0];
   int ny = ff->n[2];
   FLOAT scale = 2.0;
   FLOAT **array = (FLOAT **)malloc(nx * sizeof(FLOAT *));
   int printval;
   FILE *outfile;

   /* allocate space for 2D array */
   array[0] = (FLOAT *)malloc(nx * ny * sizeof(FLOAT));
   for (i=1; i<nx; i++)
      array[i] = array[0] + i * ny;

   /* zero the array */
   for (j=ny-1; j>=0; j--)
      for (i=0; i<nx; i++)
         array[i][j] = 0.0;

   /* fill the array */
   for (i=0; i<nx; i++)
      for (j=0; j<ny; j++) {
            array[i][j] = ff->rho[i][j]*0.0025;
      }

   // scale the array?


   /* open file for writing */
   outfile = fopen(filename,"w");
   if (outfile==NULL) {
      fprintf(stderr,"Could not open output file %s\n",filename);
      fflush(stderr);
      exit(0);
   }

   /* plot a y-plane */
   fprintf(outfile,"P2\n%d %d\n%d\n",nx,ny,255);
   for (j=ny-1; j>=0; j--) {
      for (i=0; i<nx; i++) {
         printval = (int)(256.0*array[i][j]*scale);
         if (printval<0) printval = 0;
         if (printval>255) printval = 255;
         fprintf(outfile,"%d\n",printval);
      }
   }

   /* close file */
   fclose(outfile);

   /* free memory from 2D array! */
   free(array[0]);
   free(array);

   /* if all went well, return zero */
   return(0);
}


/*
 * Put all of the particles in this cell onto the array
 */
int put_part_into_array(cell_ptr top,cell_ptr cell,FLOAT** array,int nx,int ny){

   int i,j,k;
   int xdim,ydim;
   FLOAT xstart,xsize,ystart,ysize,mult;
   particle_ptr curr;

   xdim = 0;	/* use x-values for x-dim of image */
   ydim = 2;	/* use z-values for y-dim of image */

   if (cell->has_subcells) {
      for (i=0; i<2; i++)
         for (j=0; j<2; j++)
            for (k=0; k<2; k++)
               /* put_part_into_array(top,cell,&array[0][0],nx,ny); */
               put_part_into_array(top,cell->s[i][j][k],array,nx,ny);
   } else {
      /* run through the list of elems and place each on the grid */
      // mult = nx*ny*0.05;
      mult = 255.0*sqrt(nx*ny);
      xstart = top->min[xdim];
      xsize = top->max[xdim]-xstart;
      ystart = top->min[ydim];
      ysize = top->max[ydim]-ystart;
      curr = cell->first;
      while (curr) {
         i = (int)(nx*(curr->x[xdim]-xstart)/xsize);
         j = (int)(ny*(curr->x[ydim]-xstart)/xsize);
         if (i>-1 && j>-1 && i<nx && j<ny)
            array[i][j] += mult * curr->rad;
            // array[i][j] += mult * curr->mass;
         curr = curr->next;
      }
   }
   return(0);
}


/*
 * Write a Radiance description of the particles and cell edges
 */
int write_rad(sim_ptr sim,cell_ptr top,char *filename){

   int type = 3;
   FILE *outfile;

   /* open file for writing */
   outfile = fopen(filename,"w");
   // outfile = fopen(filename,"a");
   if (outfile==NULL) {
      fprintf(stderr,"Could not open output file %s\n",filename);
      fflush(stderr);
      exit(0);
   }

   fprintf(outfile,"# Radiance scene description of dla3d run\n");
   if (type == 0) {
      /* write all cell edges, and particles as lights */
      fprintf(outfile,"void plastic edgec 0 0 5 0.2 0.2 0.2 0.0 0.0\n");
      fprintf(outfile,"void light partc 0 0 3 1000 1000 1000\n");
   } else if (type == 1) {
      /* write stars as plastic objects, and a light source overhead */
      /* fprintf(outfile,"void plastic wallc 0 0 5  0.2 0.2 0.2 0.0 0.0\n"); */
      /* fprintf(outfile,"wallc polygon floor 0 0 12  0 0 0  1 0 0  1 1 0  0 1 0\n"); */
      fprintf(outfile,"void light lightc 0 0 3  10 10 10\n");
      fprintf(outfile,"lightc polygon l1 0 0 12  0 0 2  0 1 2  1 1 2  1 0 2\n");
      fprintf(outfile,"void plastic partc 0 0 5  0.5 0.5 0.5 0.0 0.0\n");
      if (sim->bdry[2][0] == WALL) fprintf(outfile,"partc polygon floor 0 0 12 %g %g %g %g %g %g %g %g %g %g %g %g\n", top->min[0],top->min[1],top->min[2],top->max[0],top->min[1],top->min[2], top->max[0],top->max[1],top->min[2],top->min[0],top->max[1],top->min[2]);
   } else if (type == 2) {
      /* write stars as glow objects */
      fprintf(outfile,"void glow partc 0 0 4  1.0 1.0 1.0 0.0\n");
   }

   write_rad_cell(top,outfile,type);

   /* close file */
   fclose(outfile);

   /* if all went well, return zero */
   return(0);
}


/*
 * Write a Radiance description of just this cell
 */
int write_rad_cell(cell_ptr c,FILE* outfile,int type){

   int i,j,k,cnt;
   FLOAT rad;
   particle_ptr curr;

   // fprintf(outfile,"\n# cell at level %d\n",c->level);

   if (c->has_subcells) {
      /* recurse for the subcells */
      for (i=0;i<2;i++)
         for (j=0;j<2;j++)
            for (k=0;k<2;k++)
               write_rad_cell(c->s[i][j][k],outfile,type);
   } else {
      if (type == 0) {
         /* write out the edges */
         rad = 0.001;
         fprintf(outfile,"edgec cylinder e1 0 0 7 %g %g %g %g %g %g %g\n",c->min[0],c->min[1],c->min[2],c->max[0],c->min[1],c->min[2],rad);
         fprintf(outfile,"edgec cylinder e2 0 0 7 %g %g %g %g %g %g %g\n",c->min[0],c->min[1],c->max[2],c->max[0],c->min[1],c->max[2],rad);
         fprintf(outfile,"edgec cylinder e3 0 0 7 %g %g %g %g %g %g %g\n",c->min[0],c->max[1],c->min[2],c->max[0],c->max[1],c->min[2],rad);
         fprintf(outfile,"edgec cylinder e4 0 0 7 %g %g %g %g %g %g %g\n",c->min[0],c->max[1],c->max[2],c->max[0],c->max[1],c->max[2],rad);
         fprintf(outfile,"edgec cylinder e5 0 0 7 %g %g %g %g %g %g %g\n",c->min[0],c->min[1],c->min[2],c->min[0],c->max[1],c->min[2],rad);
         fprintf(outfile,"edgec cylinder e6 0 0 7 %g %g %g %g %g %g %g\n",c->min[0],c->min[1],c->max[2],c->min[0],c->max[1],c->max[2],rad);
         fprintf(outfile,"edgec cylinder e7 0 0 7 %g %g %g %g %g %g %g\n",c->max[0],c->min[1],c->min[2],c->max[0],c->max[1],c->min[2],rad);
         fprintf(outfile,"edgec cylinder e8 0 0 7 %g %g %g %g %g %g %g\n",c->max[0],c->min[1],c->max[2],c->max[0],c->max[1],c->max[2],rad);
         fprintf(outfile,"edgec cylinder e9 0 0 7 %g %g %g %g %g %g %g\n",c->min[0],c->min[1],c->min[2],c->min[0],c->min[1],c->max[2],rad);
         fprintf(outfile,"edgec cylinder e10 0 0 7 %g %g %g %g %g %g %g\n",c->min[0],c->max[1],c->min[2],c->min[0],c->max[1],c->max[2],rad);
         fprintf(outfile,"edgec cylinder e11 0 0 7 %g %g %g %g %g %g %g\n",c->max[0],c->min[1],c->min[2],c->max[0],c->min[1],c->max[2],rad);
         fprintf(outfile,"edgec cylinder e12 0 0 7 %g %g %g %g %g %g %g\n",c->max[0],c->max[1],c->min[2],c->max[0],c->max[1],c->max[2],rad);
         fflush(outfile);
      } else if (type >= 1) {
         /* don't write out the edges */
      }

      /* write out the particles */
      curr = c->first;
      cnt = 0;
      while (curr) {
         /* if (type == 0) rad = 0.001*sqrt(curr->mass); */
         /* else if (type == 1) rad = 0.5*sqrt(curr->mass); */
         /* else if (type >= 1) rad = pow(curr->mass,1./3.)/50.; */
         /* rad = pow(curr->mass,1./3.)/50.; */
         // fprintf(outfile,"partc sphere p%d 0 0 4 %g %g %g %g\n",cnt,curr->x[0],curr->x[1],curr->x[2],curr->rad);

         // if we're using cylinders instead of spheres:
         rad = 0.05*curr->rad*pow(curr->mass,1./3.);
         fprintf(outfile,"partc sphere p%d 0 0 4 %g %g %g %g\n",cnt,curr->x[0],curr->x[1],curr->x[2],rad);
         if (curr->root) {
            fprintf(outfile,"partc cylinder p%d 0 0 7 %g %g %g %g %g %g %g\n",cnt,curr->x[0],curr->x[1],curr->x[2],curr->root->x[0],curr->root->x[1],curr->root->x[2],rad);
         }

         // fflush(outfile);
         cnt++;
         // moved this outside of the loop bc of changes made 2002-09-08
         curr = curr->next;
      }
   }

   /* if all went well, return zero */
   return(0);
}


/*
 * Write a re-readable native particle3d description of the particles
 */
int write_part(cell_ptr top,char *filename){

   FILE *outfile;

   /* open file for writing */
   outfile = fopen(filename,"w");
   if (outfile==NULL) {
      fprintf(stderr,"Could not open output file %s\n",filename);
      fflush(stderr);
      exit(0);
   }

   fprintf(outfile,"# scene description of dla3d run\n");

   write_part_cell(top,outfile);

   /* close file */
   fclose(outfile);

   /* if all went well, return zero */
   return(0);
}


/*
 * Write a native particle3d description of just this cell
 */
int write_part_cell(cell_ptr c,FILE* outfile){

   int i,j,k,cnt;
   FLOAT rad;
   particle_ptr curr;

   /* fprintf(outfile,"\n# cell at level %d\n",c->level); */

   if (c->has_subcells) {
      /* recurse for the subcells */
      for (i=0;i<2;i++)
         for (j=0;j<2;j++)
            for (k=0;k<2;k++)
               write_part_cell(c->s[i][j][k],outfile);
   } else {
      /* write out the particles */
      curr = c->first;
      cnt = 0;
      while (curr) {
         // fprintf(outfile,"%g %g %g %g %g %g %g\n",curr->x[0],curr->x[1],curr->x[2],curr->rad,curr->u[0],curr->u[1],curr->u[2]);
         fprintf(outfile,"%g %g %g %g\n",curr->x[0],curr->x[1],curr->x[2],curr->rad);
         /* fflush(outfile); */
         cnt++;
         curr = curr->next;
      }
   }

   /* if all went well, return zero */
   return(0);
}

